import numpy as np
import pylab as PP
from . import fastKDE

def cumulative_integral(pdf, axes, integration_axes=None, reverse_axes=None):

    """
    Calculates the cumulative integral of the pdf, given the axis values
    
       input:
       ------
           pdf              : a PDF (presumably from the .pdf member of a fastKDE object)
           axes             : a list of PDF axes (presumably from the .axes member of a fastKDE object)
           integration_axes : the axes along which to integrate (default is all).  These have the same
                              ordering as axes in the axes input variable.
           reverse_axes     : axes along which to reverse the direction of the cumulative calculation
           
       output:
       -------
           returns an array with the same shape as PDF, but with the integral calculated cumulatively
    """

    #Set the default value for integration_axes
    if integration_axes is None:
        integration_axes = range(len(np.shape(pdf)))
    
    #Set the rank of the pdf
    rank = len(np.shape(pdf))
    
    #Set the default value for reverse_axes
    if reverse_axes is None:
        reverse_axes = []
    
    #Make sure that axes is an iterable of iterables
    try:
        axes[0][0]
    except:
        axes = [axes]
    
    #Make sure that reverse_axes is an iterable
    try:
        len(reverse_axes)
    except:
        reverse_axes = [reverse_axes]
    
    #Make sure that integration_axes is an iterable
    try:
        len(integration_axes)
    except:
        integration_axes = [integration_axes]
    
    #Copy the pdf array
    cpdf = np.array(pdf)
    
    #Loop over the axes for which we are calculating the cumulative
    for i in integration_axes:
        #Calculate the grid spacing; assume it is uniform
        dx = np.diff(axes[i])
        dx = np.concatenate([dx,[dx[-1]]])


        #Calculate the index of pdf corresponding to axis i
        iprime = rank - i - 1

        # broadcast dx to the ~same shape as cpdf
        broadcast_tuple = len(cpdf.shape)*[np.newaxis]
        broadcast_tuple[iprime] = slice(None,None,None)
        dx = dx[broadcast_tuple]

        
        #Determine if we are reversing cumulative orientation
        if i in reverse_axes:
            #Set a slice that reversese only axis i in the PDF
            reverseSlice = rank*[slice(None,None,None)]
            reverseSlice[iprime] = slice(None,None,-1)
            reverseSlice = tuple(reverseSlice)
            #Do the cumulative calculation on the reversed PDF (and reverse back to the original orientation)
            cpdf = np.cumsum(cpdf[reverseSlice]*dx,axis=iprime)[reverseSlice]
        else:
            #Do the cumulative calculation
            cpdf = np.cumsum(cpdf*dx,axis=iprime)
        
    #Return the cumulative PDF
    return cpdf


def calculate_probability_contour(pdf,axes,pvals,axis=0):
    """ Calculates the probability level below which the PDF integrates to pval. 
    
        input:
        ------
            pdf  : a conditional PDF from fastKDE
            
            axes : a list of axis values returned from fastKDE
            
            pvals : the integrated amount of PDF that should be outside a contour of constant PDF (scalar or array-like)
            
        output:
        -------
        
            an array of PDF values for which the integral of the PDF outside the PDF value should integrate to pval.
            
            Each value in the array corresponds to a value within axes[axis] (nominally, axis should be the variable
            on which the PDF is conditioned).
            
    
    """
    
    # calculate the dx values
    diff_axes = []
    for i in range(len(axes)):
        if i != axis:
            dx = np.diff(axes[i])
            dx = np.concatenate((dx,[dx[-1]]))
            diff_axes.append(dx)
            
    # create a meshgrid of dx values
    dxmesh = np.meshgrid(*diff_axes)
    # create the dx product
    dxgrid = np.prod(dxmesh,axis=0)
    # broadcast the dxgrid to the shape of the PDF
    broadcast_shape = len(axes)*[slice(None,None,None)]
    broadcast_shape[axis] = np.newaxis
    dxgrid = dxgrid[tuple(broadcast_shape[::-1])]
    
    # multiply the pdf by dx
    pdf_dx = dxgrid*pdf
    
    # reshape the PDF to unravel the summation dimensions
    new_shape = [-1,len(axes[axis])]
    pdf_unraveled = np.reshape(pdf,new_shape)
    # get the indices to sort by probability
    isort_inds = np.ma.argsort(pdf_unraveled,axis=0)
    # generate an index array suitable for use in a numpy array
    isort = ([ np.array(i) for i in zip(*isort_inds.T)], np.arange(len(axes[axis]),dtype=int)[np.newaxis,:])
    # sort by probability
    pdf_dx_unraveled = np.reshape(pdf*dxgrid,new_shape)[isort]
    
    # take the cumulative sum
    pdf_csum = np.ma.cumsum(pdf_dx_unraveled,axis=0)
    

    # find the location along each axis nearest to the given PDF value
    ind_closest = np.ma.argmin(abs(pdf_csum-pvals),axis=0)
    ind_closest = (ind_closest, np.arange(len(axes[axis]),dtype=int))

    # get the probability value at that location
    pdf_vals = np.reshape(pdf,new_shape)[isort][ind_closest]

    # return the value
    return pdf_vals

def pair_plot(var_list, \
              conditional = False, \
              fig_size = None, \
              var_names = None, \
              draw_scatter = None, \
              axis_limits = None, \
              log_scale = False, \
              auto_show = True, \
              cdf_levels = [0.1,0.25,0.5,0.75,0.9], \
              axis_lengths = None):
    """ Generate a multivariate pair plot of the input data. 
    
        input:
        ------
        
            var_list     : a list of input variables (all must be vectors of the same length) with at least two variables
              
            conditional  : flags whether to plot the bivariate PDFs as conditionals.

            var_names    : a list of strings corresponding to variable names
              
            fig_size     : The size of the figure (if none, it is set to 10,10)
            
            draw_scatter : flag whether to draw scatter plots on top of the PDFs.  If None, this turns on/off
                           automatically, depending on the number of points (1000 is the upper cutoff)
            
            axis_limits  : limits for all variables (applies to all axes)

            auto_show    : flags whether to automatically call PP.show() (no values are returned if this flag is on)
            
            cdf_levels   : a set of CDF levels (betwen 0 and 1 exclusive) for which to plot PDF contours
                           (i.e., a value of 0.5 means that 50% of the PDF falls outside the corresponding contour
                           of constant PDF that will be plotted).  If None is given, matplotlib will choose PDF levels.

            log_scale    : flags whether to use the logAxes argument for fastKDE.  If a single bool value, it applies
                           to all variables; if a list, each item in the list corresponds to a variable
            
            axis_lengths : The length of axis variables. If axis_lengths is None, then fastKDE automatically sets
                           axis lengths
                           
            
        output:
        -------

            **NOTE**: no values are returned if auto_show is True
        
            fig, axs, marginal_vals, marginal_pdfs, bivariate_pdfs   
            
                fig, axs : a matplotlib figure and axis matrix object
                
                marginal_vals : the values at which the PDFs (both marginal and bivariate) are defined
                
                marginal_pdfs : the marginal PDFs of the input variables
                
                bivariate_pdfs : the bivariate PDFs
    """
    
    # convert the list to an array
    try:
        input_var_array = np.array(var_list)
    except:
        raise ValueError("var_list could not be converted to a numpy array; are all vectors of the same length?")
        
    # get the array rank 
    rank = len(input_var_array.shape)
    if rank != 2:
        raise ValueError("input variables in var_list must be vectors")
    
    
    # get the array shape
    num_vars = input_var_array.shape[0]
    if num_vars < 2:
        raise ValueError("pair_plot requires at least two variables in var_list")
    
    if log_scale is False or log_scale is True:
        log_scale = num_vars*[log_scale]
   
    marginal_pdfs = num_vars*[None]
    marginal_vals = num_vars*[None]
    var_limits = num_vars*[None]

    if draw_scatter is None:
        if input_var_array.shape[1] > 1000:
            draw_scatter = False
        else:
            draw_scatter = True
    
    # calculate the marginals and define limits
    for n in range(num_vars):
        # calculate PDFs
        marginal_pdfs[n], marginal_vals[n] = fastKDE.pdf(input_var_array[n,:],logAxes = log_scale[n])
        
        # define axis limits
        if axis_limits is None:
            var_limits[n] = [input_var_array[n,:].min(),input_var_array[n,:].max()]
        else:
            var_limits[n] = axis_limits
        
        
    #bivariate_pdfs = num_vars*[num_vars*[None]]
    bivariate_pdfs = {}
    
    # calculate the bivariate PDFs
    for n1 in range(num_vars):
        for n2 in range(n1, num_vars):
            if n1 != n2:
                if not conditional:
                    bivariate_pdfs[n1,n2], _ = fastKDE.pdf(input_var_array[n1,:], input_var_array[n2,:],logAxes = [log_scale[n1],log_scale[n2]])
                else:
                    bivariate_pdfs[n1,n2], _ = fastKDE.conditional(input_var_array[n2,:], input_var_array[n1,:],logAxes = [log_scale[n1],log_scale[n2]])
                    bivariate_pdfs[n2,n1], _ = fastKDE.conditional(input_var_array[n1,:], input_var_array[n2,:],logAxes = [log_scale[n2],log_scale[n1]])

            
            
    # set variable labels if needed
    if var_names is None:
        var_names = ['var {}'.format(n) for n in range(num_vars)]
        
    # set the figure size if needed
    if fig_size is None:
        fig_size = (10,10)
        
    # create the figure
    fig, axs = PP.subplots(num_vars,num_vars,figsize=fig_size)
    
    # plot the marginals
    for n in range(num_vars):
        axs[n,n].plot(marginal_vals[n],marginal_pdfs[n])
        axs[n,n].set_xlim(var_limits[n])
        axs[n,n].tick_params(labelleft = False)    
        if n < num_vars - 1:
            axs[n,n].tick_params(labelbottom = False)    
        else:
            axs[n,n].set_xlabel(var_names[n])
            
        if log_scale[n]:
            axs[n,n].set_xscale('log')
        
            
    # plot the bivariate PDFs
    for n1 in range(num_vars):
        for n2 in range(n1, num_vars):
            if n1 != n2:
                
                if not conditional:
                    # get the PDF levels
                    if cdf_levels is None:
                        levels = None
                    else:
                        levels = np.sort(np.array([calculate_probability_contour(  bivariate_pdfs[n1,n2][...,np.newaxis],\
                                                                           [[0],marginal_vals[n1],marginal_vals[n2]],\
                                                                           c) 
                                                   for c in cdf_levels])).squeeze()
                    
                    # plot the PDF
                    pdf_to_plot = bivariate_pdfs[n1,n2]
                    if any([log_scale[n1],log_scale[n2]]):
                        pdf_to_plot = np.ma.log(np.ma.masked_less_equal(bivariate_pdfs[n1,n2],0))
                        levels = np.log(levels)
                else:
                    levels = cdf_levels
                    pdf_to_plot = cumulative_integral(bivariate_pdfs[n1,n2],[marginal_vals[n1],marginal_vals[n2]],integration_axes=1)
                    # mask CDF parts that don't normalize to something close to 1
                    pdf_to_plot = np.ma.masked_where(np.logical_not(np.isclose(np.ma.ones(pdf_to_plot.shape)*pdf_to_plot[-1,:][np.newaxis,:],1.0)),pdf_to_plot)

                
                axs[n2,n1].contour(marginal_vals[n1],marginal_vals[n2],pdf_to_plot,levels=levels)
                    
                if log_scale[n1]:
                    axs[n2,n1].set_xscale('log')
                if log_scale[n2]:
                    axs[n2,n1].set_yscale('log')

                # plot the scatter
                if draw_scatter:
                    axs[n2,n1].plot(input_var_array[n1,:],input_var_array[n2,:],'k.',alpha = 0.3)
                    
                # set axis limits
                axs[n2,n1].set_xlim(var_limits[n1])
                axs[n2,n1].set_ylim(var_limits[n2])

                if not conditional:
                    # turn of axes for the other part of the triangle
                    axs[n1,n2].axis('off')
                else:

                    levels = cdf_levels
                    pdf_to_plot = cumulative_integral(bivariate_pdfs[n2,n1],[marginal_vals[n2],marginal_vals[n1]],integration_axes=1)
                    # mask CDF parts that don't normalize to something close to 1
                    pdf_to_plot = np.ma.masked_where(np.logical_not(np.isclose(np.ma.ones(pdf_to_plot.shape)*pdf_to_plot[-1,:][np.newaxis,:],1.0)),pdf_to_plot)

                    axs[n1,n2].contour(marginal_vals[n2],marginal_vals[n1],pdf_to_plot,levels=levels)
                        
                    if log_scale[n2]:
                        axs[n1,n2].set_xscale('log')
                    if log_scale[n1]:
                        axs[n1,n2].set_yscale('log')

                    # plot the scatter
                    if draw_scatter:
                        axs[n1,n2].plot(input_var_array[n2,:],input_var_array[n1,:],'k.',alpha = 0.3)
                        
                    # set axis limits
                    axs[n1,n2].set_xlim(var_limits[n2])
                    axs[n1,n2].set_ylim(var_limits[n1])


        
        
    
    # turn off axis limits and set axis labels
    for n1 in range(num_vars):
        for n2 in range(n1, num_vars):
            if n1 != n2:
                if n1 == 0: 
                    axs[n2,n1].set_ylabel(var_names[n2])
                    axs[n2,n1].tick_params(labelleft = True)    
                else:
                    axs[n2,n1].tick_params(labelleft = False)    
                    
                if n2 == num_vars - 1:
                    axs[n2,n1].set_xlabel(var_names[n1])
                    axs[n2,n1].tick_params(labelbottom = True)    
                else:
                    axs[n2,n1].tick_params(labelbottom = False)    

                if conditional:
                    axs[n1,n2].tick_params(labelleft = False)    
                    axs[n1,n2].tick_params(labelbottom = False)    
                    if n2 == num_vars - 1:
                        axs[n1,n2].tick_params(labelright = True)    
                    
    if auto_show:
        PP.show()
        return
    
    return fig, axs, marginal_vals, marginal_pdfs, bivariate_pdfs






import numpy as np
cimport numpy as np
cimport cython
import numpy.fft as fft
import cython.parallel as cpy

cdef extern from "math.h" nogil:
    double floor(double x)
    double exp(double x)
    double cos(double x)
    double sin(double x)

cdef double complex cexp(double complex x) nogil:
    cdef double complex result
    cdef double re
    cdef double im
    re = x.real
    im = x.imag
    result = exp(re)*(cos(im) + 1.0j*sin(im))
    return result

#*******************************************************************************
#*******************************************************************************
#************************** nuifft *********************************************
#*******************************************************************************
#*******************************************************************************
@cython.boundscheck(False)
cpdef np.ndarray[double complex] nuifft( \
                            np.float_t [:,:] abscissas, \
                            np.complex128_t [:] ordinates, \
                            np.float_t [:,:] frequencyGrids, \
                            np.float_t missingFreqVal = -1e20, \
                            int precision = 2, \
                            int beVerbose = 0):
    """Approximates the unnormalized direct Fourier transform of abscissa, ordinate pairs
       using the non-uniform FFT method.
        
        input:
        ------

            abscissas   : abscissa values.
                          A numpy array of shape (ndimensions,npoints)
                          (assumed to be real)

            ordinates   : ordinates values.
                          A numpy array of shape (npoints)

            frequencyGrids : The frequency grids on which to calcualte the DFT
                             A masked numpy array of shape (ndimensions,ntmax), where ntmax
                             is the length of the longest frequency grid.  

            missingFreqVal : A value indicating missing frequency values.  This is used to 
                             allow each dimension to have different sized frequency spaces;
                             dimensions with smaller frequency spaces (than ntmax) should be 
                             padded at the end with missingFreqVal.

            precision   : Sets the precision of the approximation.  1 for floating point
                          and 2 for double precision accuracy.

            beVerbose   : Flags whether to print to STDOUT as the method progresses
                          (int 0=don't print 1=print)

        output:
        -------

            The DFT of abscissa/ordinate pairs.  Calculated in one dimension as:

            DFT[t] = sum( [ a[i] * exp(j * x[i] * t ) for i in range(N) ] ) 

            for each of the t points in frequencyGrids (and where j = sqrt(-1)).
    
    """

    cdef int numDimensions
    cdef int numDataPoints
    cdef int ntMax

    #*******************************
    # Get variable dimensionalities
    # and do consistency checks
    #*******************************
    vprint("Checking dimensionalities and arguments",beVerbose)
    #Get the shape of abscissas 
    try:
        numDimensions = np.shape(abscissas)[0]
        numDataPoints = np.shape(abscissas)[1]
    except:
        raise ValueError,"Could not determine shape of abscissas"

    #Check ordinates
    try:
        ordShape = np.shape(ordinates)[0]
    except:
        raise ValueError,"Could not determine shape of ordinates"
        
    if ordShape != numDataPoints:
            raise ValueError, "Incompatible shapes for ordinates and abscissas"

    #Check frequencyGrids
    try:
        freqShape = np.shape(frequencyGrids)
    except:
        raise ValueError,"Could not determine shape of ordinates"

    if freqShape[0] != numDimensions:
            raise ValueError, "Incompatible shapes for abscissas and frequencyGrids"

    #Get the max number of frequency points
    ntMax = freqShape[1]


    #********************************************
    # Calculate the size of the frequency spaces
    #********************************************
    vprint("Getting the size of the frequency spaces",beVerbose)
    cdef int n,t,iNotMissing
    cdef np.int_t [:] frequencySizes = np.zeros([numDimensions],dtype=np.int)

    for n in range(numDimensions):
        iNotMissing = 0
        for t in range(ntMax):
            if frequencyGrids[n,t] != missingFreqVal:
                iNotMissing += 1
        if iNotMissing != 0:
            frequencySizes[n] = <np.int_t>iNotMissing
        else:
            raise ValueError,"Some frequencies in frequencyGrids have no valid points"

    #Get the total size of the frequency space
    cdef np.int_t freqSpaceSize = np.prod(frequencySizes)



    cdef np.float_t [:,:] \
            abscissaGrids = missingFreqVal*np.ones([numDimensions,ntMax])
    #******************************
    # Calculate the abscissa grids
    #******************************
    vprint("Creating the convolution grid",beVerbose)
    for n in range(numDimensions):
        xdum = calcXfromT(frequencyGrids[n,:frequencySizes[n]])
        for t in range(frequencySizes[n]):
            abscissaGrids[n,t] = xdum[t]


    #*******************************
    # Define convolution parameters 
    #*******************************
    vprint("Initializing the convolution",beVerbose)
    cdef np.float_t tau,fourTau
    cdef long nspread
    cdef long nspreadhalf, hyperSlabSize

    #Set the parameters that control the precision of the nuFFT
    if precision == 1:
        tau = 0.5993
        nspread = 10
    else:
        tau = 1.5629
        nspread = 28
    nspreadhalf = nspread/2
    fourTau = 4*tau
    hyperSlabSize = nspread**numDimensions

    #get the shape of a hyperslab
    #cdef np.ndarray[np.int_t,ndim=1] hyperSlabShape = nspread*np.ones([numDimensions],dtype=np.int)
    cdef np.int_t [:] hyperSlabShape = nspread*np.ones([numDimensions],dtype=np.int)
    vprint("\tconvolution hyperslab shape: {}".format(hyperSlabShape),beVerbose)

    #Calculate the quantities necessary for estimating x-indices
    cdef np.float_t [:]  \
            xmins = np.array([ abscissaGrids[n,0] for n in range(numDimensions) ])
    cdef np.float_t [:]  \
            deltaxs = np.array([ abscissaGrids[n,1] - abscissaGrids[n,0] for n in range(numDimensions) ])

    #Inialize worker terms for the convolution
    #cdef np.ndarray[long,ndim=1] mvec = np.zeros([numDimensions],dtype=np.int)
    cdef np.int_t [:] mvec = np.zeros([numDimensions],dtype=np.int)
    cdef np.int_t [:] m0vec = np.zeros([numDimensions],dtype=np.int)
    cdef np.float_t [:] mprimevec = np.zeros([numDimensions])
    cdef np.float_t mprime = 0.0
    cdef double complex gaussTerm = 0.0
    cdef np.float_t gaussArg = 0.0
    cdef int j, v, i, k
    i = j = v = k = 0

    #Initialize the (raveled) data convolution
    cdef double complex [:] convolvedData = np.zeros([freqSpaceSize],dtype=np.complex128)


    #***********************************
    # Convolve the data with a gaussian
    #***********************************
    vprint("Performing the convolution",beVerbose)
    with nogil:
        for j in range(numDataPoints):
            
            #Transform the coordinates of our current abscissas value
            #into the coordinate system of a hyperslab centered on the nearest
            #point (nearest in the floor() sense)
            for v in range(numDimensions):
                #Transform the current abscissa value into an approximate
                #grid value
                mprime = (abscissas[v,j] - xmins[v])/deltaxs[v]
                #Calculate the nearest (in the floor sense) grid abscissa
                m0vec[v] = <long> floor(mprime)
                #Finish transforming the absicssa value to hyperslab coordinates
                mprimevec[v] = mprime - m0vec[v] + nspreadhalf

            #Cycle over all the hyperslab points and calculate the gaussian convolution
            #of the current data point only on points in the hyperslab
            for i in range(hyperSlabSize):
                #Get the indices (mvec) of the current point (i) in our hyperslab
                unravelIndex(i,hyperSlabShape,mvec,numDimensions)

                gaussArg = 0.0
                for v in range(numDimensions):
                    #Calculate the distance between the data point and the current
                    #hyperslab point
                    gaussArg = gaussArg + (<np.float_t>mvec[v] - mprimevec[v])**2
                    #Transform from hyperslab coordinates back to full space coordinates
                    mvec[v] = (mvec[v] + m0vec[v]) - nspreadhalf
                #Calculate the gaussian of this distance
                gaussTerm = ordinates[j]*(<double complex> exp(-gaussArg/fourTau))

                #Calculate the flattened array index of the current point
                k = ravelIndex(frequencySizes,mvec,numDimensions)

                #Add the gaussian term to this point (only if it is within our domain
                if k >= 0 and k < freqSpaceSize:
                    convolvedData[k] = convolvedData[k] + gaussTerm

    #Take its FFT
    #( ifftshift is used to reorder the kde array such that 0 is the lowest corner 
    #  first index] of the array, as required by ifft)
    #And fftshift is used to put the zero-frequency in the center of the array
    #reshape is used to form a proper n-dimensional array from the vector of convolved data
    vprint("Raveling and doing FFT on convolved data",beVerbose)
    convolvedFFT = fft.fftshift(fft.ifftn(fft.ifftshift(np.reshape(convolvedData,frequencySizes))))
    vprint("\tshape(convolvedFFT) = {}".format(np.shape(convolvedFFT)),beVerbose)

    vprint("Initializing the deconvolution",beVerbose)

    #Ravel the convolved data
    cdef double complex [:] convolvedFFTRaveled = convolvedFFT.ravel()

    #Pre-declare and allocate a raveled form of the DFT
    cdef double complex [:] DFT = np.zeros([freqSpaceSize],dtype=np.complex128)
    #Pre declare a dimension index vector
    cdef np.int_t [:] dimInds = np.zeros([numDimensions],dtype=np.int)

    #Deconvolve the FFT (divide by the FFT of the gaussian) to obtain the DFT estimate
    vprint("Deconvolving the Fourier transformed data",beVerbose)
    with nogil:
        for i in range(freqSpaceSize):
            unravelIndex(i,frequencySizes,dimInds,numDimensions)
            gaussArg = 0.0
            for v in range(numDimensions):
               gaussArg = gaussArg + (frequencyGrids[v,dimInds[v]]*deltaxs[v])**2
            DFT[i] = convolvedFFTRaveled[i] * cexp(tau*gaussArg)


    #Reshape the DFT to an array
    vprint("Reshaping and returning.",beVerbose)
    return np.reshape(DFT,tuple(frequencySizes))


#*******************************************************************************
#*******************************************************************************
#************************** idft ***********************************************
#*******************************************************************************
#*******************************************************************************
@cython.boundscheck(False)
cpdef np.ndarray[double complex] idft( \
                            np.float_t [:,:] abscissas, \
                            np.complex128_t [:] ordinates, \
                            np.float_t [:,:] frequencyGrids, \
                            np.float_t missingFreqVal = -1e20, \
                            beVerbose = False):
    """Calculates the unnormalized direct Fourier transform of abscissa, ordinate pairs
        
        input:
        ------

            abscissas   : abscissa values.
                          A numpy array of shape (ndimensions,npoints)
                          (assumed to be real)

            ordinates   : ordinates values.
                          A numpy array of shape (npoints)

            frequencyGrids : The frequency grids on which to calcualte the DFT
                             A masked numpy array of shape (ndimensions,ntmax), where ntmax
                             is the length of the longest frequency grid.  

            missingFreqVal : A value indicating missing frequency values.  This is used to 
                             allow each dimension to have different sized frequency spaces;
                             dimensions with smaller frequency spaces (than ntmax) should be 
                             padded at the end with missingFreqVal.

            beVerbose   : Flags whether to print to STDOUT as the method progresses
                          (int 0=don't print 1=print)
        output:
        -------

            The DFT of abscissa/ordinate pairs.  Calculated in one dimension as:

            DFT[t] = sum( [ a[i] * exp(j * x[i] * t ) for i in range(N) ] ) 

            for each of the t points in frequencyGrids (and where j = sqrt(-1)).
    
    """

    cdef int numDimensions
    cdef int numDataPoints
    cdef int ntMax

    #*******************************
    # Get variable dimensionalities
    # and do consistency checks
    #*******************************
    #Get the shape of abscissas 
    vprint("Checking dimensionalities and arguments",beVerbose)
    try:
        numDimensions = np.shape(abscissas)[0]
        numDataPoints = np.shape(abscissas)[1]
    except:
        raise ValueError,"Could not determine shape of abscissas"

    #Check ordinates
    try:
        ordShape = np.shape(ordinates)[0]
    except:
        raise ValueError,"Could not determine shape of ordinates"
        
    if ordShape != numDataPoints:
            raise ValueError, "Incompatible shapes for ordinates and abscissas"

    #Check frequencyGrids
    try:
        freqShape = np.shape(frequencyGrids)
    except:
        raise ValueError,"Could not determine shape of ordinates"

    if freqShape[0] != numDimensions:
            raise ValueError, "Incompatible shapes for abscissas and frequencyGrids"

    #Get the max number of frequency points
    ntMax = freqShape[1]


    #********************************************
    # Calculate the size of the frequency spaces
    #********************************************
    vprint("Getting the size of the frequency spaces",beVerbose)
    cdef int n,t,iNotMissing
    cdef np.int_t [:] frequencySizes = np.zeros([numDimensions],dtype=np.int)

    for n in range(numDimensions):
        iNotMissing = 0
        for t in range(ntMax):
            if frequencyGrids[n,t] != missingFreqVal:
                iNotMissing += 1
        if iNotMissing != 0:
            frequencySizes[n] = <np.int_t>iNotMissing
        else:
            raise ValueError,"Some frequencies in frequencyGrids have no valid points"

    #Get the total size of the frequency space
    cdef np.int_t freqSpaceSize = np.prod(frequencySizes)

    #Pre-declare and allocate a raveled form of the DFT
    cdef double complex [:] DFT = np.zeros([freqSpaceSize],dtype=np.complex128)

    cdef int i,k

    #Pre declare a dimension index vector
    cdef np.int_t [:] dimInds = np.zeros([numDimensions],dtype=np.int)

    cdef double complex myDFT
    cdef double expArg

    cdef double complex dftConst = 1.0j #* <double complex> np.pi


    vprint("Calculting the DFT",beVerbose)
    with nogil:
        for i in range(freqSpaceSize):
            unravelIndex(i,frequencySizes,dimInds,numDimensions)

            myDFT = 0.0 + 0.0j 

            for j in range(numDataPoints):
                expArg = 0.0

                for k in range(numDimensions):
                    expArg = expArg +(abscissas[k,j] * frequencyGrids[k,dimInds[k]])

                myDFT = myDFT + ordinates[j]*cexp(dftConst * <double complex> expArg)

            DFT[i] = myDFT

    return np.reshape(DFT,tuple(frequencySizes))

#*******************************************************************************
#*******************************************************************************
#********************* unravelIndex() ******************************************
#*******************************************************************************
#*******************************************************************************
@cython.boundscheck(False)
cdef inline int unravelIndex( \
                   np.int_t i, \
                   np.int_t [:] frequencySizes, \
                   np.int_t [:] dimInds, \
                   np.int_t ndims) nogil:
    """Takes the 1D index i of a raveled variable of shape frequencySizes and returns
    an array of the unraveled indices."""

    cdef np.int_t n,nd,iDum
    cdef np.int_t hyperSize

    hyperSize = 1
    for n in range(1,ndims):
        hyperSize *= frequencySizes[n]
        
    iDum = i
    for n in range(ndims):
        dimInds[n] = <np.int_t> floor((<double> iDum)/(<double> hyperSize))
        iDum -= dimInds[n]*hyperSize
        if n < (ndims-1):
            hyperSize /= frequencySizes[n+1]
        else:
            hyperSize = 1

    return 0

#*******************************************************************************
#*******************************************************************************
#********************* ravelIndex() ********************************************
#*******************************************************************************
#*******************************************************************************
@cython.boundscheck(False)
cdef inline int ravelIndex( \
                   np.int_t [:] frequencySizes, \
                   np.int_t [:] dimInds, \
                   np.int_t ndims) nogil:
    """Calculates the 1D index i of a raveled variable of shape frequencySizes, given
    an array of the unraveled indices."""

    cdef np.int_t n
    cdef np.int_t hyperSize
    cdef np.int_t i

    hyperSize = 1
    i = 0
    for n in range(ndims-1,-1,-1):
        i += hyperSize*dimInds[n]
        hyperSize *= frequencySizes[n]
    return i

#*****************************************************************************
#*****************************************************************************
#******************* Frequency/real-space conversions ************************
#*****************************************************************************
#*****************************************************************************
def calcXfromT(tpoints):
  """Calculates real space points given a set of hermetian frequency points. """
  #Use fftfreq to produce a set of frequencies that correspond to the fourier
  #transform of a signal on the tpoints points
  deltaT = tpoints[1] - tpoints[0]
  return  fft.fftshift(fft.fftfreq(len(tpoints),deltaT/(2*np.pi)))

def calcTfromX(xpoints):
  """Calculates frequency points given a signal in real space. """
  #Use fftfreq to produce a set of frequencies that correspond to the fourier
  #transform of a signal on the x points
  deltaX = xpoints[1] - xpoints[0]
  return fft.fftshift(fft.fftfreq(len(xpoints),deltaX/(2*np.pi)))

def vprint(msg,beVerbose):
    """Prints only if beVerbose is True"""
    if beVerbose:
        print(msg)


#*******************************************************************************
#*******************************************************************************
#************************** dft_points *****************************************
#*******************************************************************************
#*******************************************************************************
@cython.boundscheck(False)
cpdef np.ndarray[double complex] dft_points( \
                            np.float_t [:,:] frequencyGrids, \
                            np.complex128_t [:] ordinates, \
                            np.float_t [:,:] outputPoints, \
                            np.float_t missingFreqVal = -1e20, \
                            beVerbose = False):
    """Calculates the unnormalized direct inverse Fourier transform of abscissa, ordinate pairs
        
        input:
        ------

            frequencyGrids   : abscissa values.
                          A numpy array of shape (ndimensions,npoints)
                          (assumed to be real)

            ordinates   : ordinates values.
                          A numpy array of shape (npoints)

            outputPoints :  The real-space points at which to calcualte the DFT
                             A masked numpy array of shape (ndimensions,noutputpoints). 


            beVerbose   : Flags whether to print to STDOUT as the method progresses
                          (int 0=don't print 1=print)
        output:
        -------

            The iDFT of abscissa/ordinate pairs.  Calculated in one dimension as:

            iDFT[x] = sum( [ a[i] * exp(-j * x * t[i] ) for i in range(N) ] ) 

            for each of the x points in output_points (and where j = sqrt(-1)).
    
    """

    cdef int numDimensions
    cdef int numDataPoints
    cdef int numOutputPoints

    #*******************************
    # Get variable dimensionalities
    # and do consistency checks
    #*******************************
    #Get the shape of frequencyGrids 
    vprint("Checking dimensionalities and arguments",beVerbose)
    try:
        numDimensions = np.shape(frequencyGrids)[0]
    except:
        raise ValueError,"Could not determine shape of frequencyGrids"

    #Check ordinates
    try:
        ordShape = np.shape(ordinates)[0]
    except:
        raise ValueError,"Could not determine shape of ordinates"
        
    #Check outputPoints
    try:
        outputShape = np.shape(outputPoints)
    except:
        raise ValueError,"Could not determine shape of outputPoints"

    if outputShape[0] != numDimensions:
            raise ValueError, "Incompatible shapes for frequencyGrids and outputPoints"

    #Get the number output data points
    numOutputPoints = outputShape[1]

    #Pre-declare and allocate the iDFT
    cdef np.float_t [:] iDFT = np.zeros([numOutputPoints],dtype=float)

    cdef int i,k,n

    cdef double myiDFT
    cdef double expArg

    cdef double complex idftConst = -1.0j


    #Get the max number of frequency points
    cdef int ntMax
    ntMax = np.shape(frequencyGrids)[1]

    #********************************************
    # Calculate the size of the frequency spaces
    #********************************************
    vprint("Getting the size of the frequency spaces",beVerbose)
    cdef int t,iNotMissing
    cdef np.int_t [:] frequencySizes = np.zeros([numDimensions],dtype=np.int)

    for n in range(numDimensions):
        iNotMissing = 0
        for t in range(ntMax):
            if frequencyGrids[n,t] != missingFreqVal:
                iNotMissing += 1
        if iNotMissing != 0:
            frequencySizes[n] = <np.int_t>iNotMissing
        else:
            raise ValueError,"Some frequencies in frequencyGrids have no valid points"

    #Get the total size of the frequency space
    cdef np.int_t freqSpaceSize = np.prod(frequencySizes)

    #Pre-declare and allocate a raveled form of the DFT
    cdef double complex [:] DFT = np.zeros([freqSpaceSize],dtype=np.complex128)

    #Pre declare a dimension index vector
    cdef np.int_t [:] dimInds = np.zeros([numDimensions],dtype=np.int)

    vprint("Calculating the DFT",beVerbose)
    with nogil:
        for i in range(numOutputPoints):
            myiDFT = 0.0 


            for j in range(freqSpaceSize):
                unravelIndex(j,frequencySizes,dimInds,numDimensions)
                expArg = 0.0

                for k in range(numDimensions):

                    expArg = expArg +(frequencyGrids[k,dimInds[k]] * outputPoints[k,i])

                #Calculate the flattened array index of the current point
                myiDFT = myiDFT + (ordinates[j]*cexp(idftConst * <double complex> expArg)).real

            iDFT[i] = myiDFT

    return np.array(iDFT,copy=True)
